unit PDSL.Expressions;

interface

uses
  System.SysUtils, Generics.Collections;

type
  EExprException = class(Exception);
  IExpression = interface;

  IEnvironment = interface
    function GetValue(const AName: string): IExpression;
    function SetValue(const AName: string; AExpr: IExpression): IEnvironment;
  end;

  IExpression = interface
    function GetAsString: string;

    function Evaluate: IExpression; overload;
    function Evaluate(Env: IEnvironment): IExpression; overload;

    property AsString: string read GetAsString;
  end;

  IHasValue = interface
    ['{567A6313-3ABE-4620-9560-64F93BC4979A}']
    function GetValue: Variant;
    property Value: Variant read GetValue;
  end;

  IClosure = interface
    ['{BB51A47B-B876-43A9-AA81-2C3124299FC0}']
    function GetEnv: IEnvironment;
    procedure SetEnv(Value: IEnvironment);

    function EvaluateClosure(AParamValue: IExpression): IExpression;

    property Env: IEnvironment read GetEnv write SetEnv;
  end;

  TExpression = class abstract(TInterfacedObject, IExpression)
  protected
    FExprs: TList<IExpression>;
    function GetAsString: string; virtual;
  public
    constructor Create;
    destructor Destroy; override;

    function Evaluate: IExpression; overload; virtual;
    function Evaluate(Env: IEnvironment): IExpression; overload; virtual; abstract;

    property AsString: string read GetAsString;
  end;

  TNumber = class(TExpression, IHasValue)
  private
    FValue: Integer;
    function GetValue: Variant;
  protected
    function GetAsString: string; override;
  public
    constructor Create(AValue: Integer); reintroduce;

    function Evaluate(Env: IEnvironment): IExpression; override;

    property Value: Variant read GetValue;
  end;

  TVariable = class(TExpression)
  private
    FName: string;
  protected
    function GetAsString: string; override;
  public
    constructor Create(AName: string); reintroduce;
    function Evaluate(Env: IEnvironment): IExpression; override;
  end;

  TAdd = class(TExpression)
  public
    constructor Create(AValue1, AValue2: IExpression); reintroduce;

    function Evaluate(Env: IEnvironment): IExpression; override;
  end;

  TSub = class(TExpression)
  public
    constructor Create(AValue1, AValue2: IExpression); reintroduce;

    function Evaluate(Env: IEnvironment): IExpression; override;
  end;

  TMul = class(TExpression)
  public
    constructor Create(AValue1, AValue2: IExpression); reintroduce;

    function Evaluate(Env: IEnvironment): IExpression; override;
  end;

  TEquals = class(TExpression)
  public
    constructor Create(AValue1, AValue2: IExpression); reintroduce;

    function Evaluate(Env: IEnvironment): IExpression; override;
  end;

  TIfThenElse = class(TExpression)
  public
    constructor Create(AExpr1, AExpr2, AExpr3: IExpression); reintroduce;

    function Evaluate(Env: IEnvironment): IExpression; override;
  end;

  TLet = class(TExpression)
  private
    FVarName: string;
  protected
    function GetAsString: string; override;
  public
    constructor Create(const AVarName: string; AVarValue, ABody: IExpression); reintroduce;

    function Evaluate(Env: IEnvironment): IExpression; override;
  end;

  TClosure = class(TExpression, IClosure)
  private
    FEnv: IEnvironment;
    FParamName: string;

    function GetEnv: IEnvironment;
    procedure SetEnv(Value: IEnvironment);
  public
    constructor Create(AFunc: IExpression; AEnv: IEnvironment; const AParamName: string); reintroduce;

    function EvaluateClosure(AParamValue: IExpression): IExpression;

    property Env: IEnvironment read GetEnv write SetEnv;
  end;

  TDefineFunc = class(TExpression)
  private
    FParamName: string;
  protected
    function GetAsString: string; override;
  public
    constructor Create(const AParamName: string; ABody: IExpression); reintroduce;

    function Evaluate(Env: IEnvironment): IExpression; override;
  end;

  TCallFunc = class(TExpression)
  private
    FFuncName: string;
  protected
    function GetAsString: string; override;
  public
    constructor Create(const AFuncName: string; AParamValue: IExpression); reintroduce;

    function Evaluate(Env: IEnvironment): IExpression; override;
  end;

implementation

uses
  PDSL.Environment;

{ TExpression }

constructor TExpression.Create;
begin
  inherited;
  FExprs := TList<IExpression>.Create;
end;

destructor TExpression.Destroy;
begin
  FExprs.Free;
  inherited;
end;

function TExpression.Evaluate: IExpression;
begin
  Result := Evaluate(TEnvironment.Create);
end;

function TExpression.GetAsString: string;
var
  Expr: IExpression;
begin
  Result := '(' + Self.ClassName;
  for Expr in FExprs do
    Result := Result + ' ' + Expr.AsString;
  Result := Result + ')';
end;

{ TNumber }

constructor TNumber.Create(AValue: Integer);
begin
  inherited Create;
  FValue := AValue;
end;

function TNumber.Evaluate(Env: IEnvironment): IExpression;
begin
  Result := Self;
end;

function TNumber.GetAsString: string;
begin
  Result := Format('(%s %d)', [Self.ClassName, FValue]);
end;

function TNumber.GetValue: Variant;
begin
  Result := FValue;
end;

{ TVariable }

constructor TVariable.Create(AName: string);
begin
  inherited Create;
  FName := AName;
end;

function TVariable.Evaluate(Env: IEnvironment): IExpression;
begin
  Result := Env.GetValue(FName);
end;

function TVariable.GetAsString: string;
begin
  Result := Format('(%s %s)', [Self.ClassName, FName]);
end;

{ TAdd }

constructor TAdd.Create(AValue1, AValue2: IExpression);
begin
  inherited Create;
  FExprs.Add(AValue1);
  FExprs.Add(AValue2);
end;

function TAdd.Evaluate(Env: IEnvironment): IExpression;
var
  Expr1, Expr2: IExpression;
  Val1, Val2: IHasValue;
begin
  Expr1 := FExprs[0].Evaluate(Env);
  Expr2 := FExprs[1].Evaluate(Env);

  if Supports(Expr1, IHasValue, Val1) and Supports(Expr2, IHasValue, Val2) then
    Result := TNumber.Create(Val1.Value + Val2.Value)
  else
    raise EExprException.Create('Invalid expression applied to TAdd');
end;

{ TSub }

constructor TSub.Create(AValue1, AValue2: IExpression);
begin
  inherited Create;
  FExprs.Add(AValue1);
  FExprs.Add(AValue2);
end;

function TSub.Evaluate(Env: IEnvironment): IExpression;
var
  Expr1, Expr2: IExpression;
  Val1, Val2: IHasValue;
begin
  Expr1 := FExprs[0].Evaluate(Env);
  Expr2 := FExprs[1].Evaluate(Env);

  if Supports(Expr1, IHasValue, Val1) and Supports(Expr2, IHasValue, Val2) then
    Result := TNumber.Create(Val1.Value - Val2.Value)
  else
    raise EExprException.Create('Invalid expression applied to TSub');
end;

{ TMul }

constructor TMul.Create(AValue1, AValue2: IExpression);
begin
  inherited Create;
  FExprs.Add(AValue1);
  FExprs.Add(AValue2);
end;

function TMul.Evaluate(Env: IEnvironment): IExpression;
var
  Expr1, Expr2: IExpression;
  Val1, Val2: IHasValue;
begin
  Expr1 := FExprs[0].Evaluate(Env);
  Expr2 := FExprs[1].Evaluate(Env);

  if Supports(Expr1, IHasValue, Val1) and Supports(Expr2, IHasValue, Val2) then
    Result := TNumber.Create(Val1.Value * Val2.Value)
  else
    raise EExprException.Create('Invalid expression applied to TMul');
end;

{ TEquals }

constructor TEquals.Create(AValue1, AValue2: IExpression);
begin
  inherited Create;
  FExprs.Add(AValue1);
  FExprs.Add(AValue2);
end;

function TEquals.Evaluate(Env: IEnvironment): IExpression;
var
  Expr1, Expr2: IExpression;
  Val1, Val2: IHasValue;
begin
  Expr1 := FExprs[0].Evaluate(Env);
  Expr2 := FExprs[1].Evaluate(Env);

  if Supports(Expr1, IHasValue, Val1) and Supports(Expr2, IHasValue, Val2) then
  begin
    if Val1.Value = Val2.Value then
      Result := TNumber.Create(1)
    else
      Result := TNumber.Create(0);
  end
  else
    raise EExprException.Create('Invalid expression applied to TMul');
end;

{ TIfThenElse }

constructor TIfThenElse.Create(AExpr1, AExpr2, AExpr3: IExpression);
begin
  inherited Create;
  FExprs.Add(AExpr1);
  FExprs.Add(AExpr2);
  FExprs.Add(AExpr3);
end;

function TIfThenElse.Evaluate(Env: IEnvironment): IExpression;
var
  Expr1: IExpression;
  Val1: IHasValue;
begin
  Expr1 := FExprs[0].Evaluate(Env);

  if Supports(Expr1, IHasValue, Val1) then
  begin
    if Val1.Value > 0 then
      Result := FExprs[1].Evaluate(Env)
    else
      Result := FExprs[2].Evaluate(Env);
  end
  else
    raise EExprException.Create('Invalid expression applied to TIfThenElse');
end;

{ TLet }

constructor TLet.Create(const AVarName: string; AVarValue, ABody: IExpression);
begin
  inherited Create;
  FVarName := AVarName;
  FExprs.Add(AVarValue);
  FExprs.Add(ABody);
end;

function TLet.Evaluate(Env: IEnvironment): IExpression;
var
  VarValue: IExpression;
  Closure: IClosure;
begin
  VarValue := FExprs[0].Evaluate(Env);
  if Supports(VarValue, IClosure, Closure) then
    Closure.Env := Closure.Env.SetValue(FVarName, VarValue);

  Result := FExprs[1].Evaluate(Env.SetValue(FVarName, VarValue));
end;

function TLet.GetAsString: string;
begin
  Result := Format('(%s [%s %s] %s)', [Self.ClassName, FVarName,
    FExprs[0].AsString, FExprs[1].AsString]);
end;

{ TClosure }

constructor TClosure.Create(AFunc: IExpression; AEnv: IEnvironment; const AParamName: string);
begin
  inherited Create;
  FEnv := AEnv;
  FParamName := AParamName;
  FExprs.Add(AFunc);
end;

function TClosure.EvaluateClosure(AParamValue: IExpression): IExpression;
begin
  Result := FExprs[0].Evaluate(FEnv.SetValue(FParamName, AParamValue));
end;

function TClosure.GetEnv: IEnvironment;
begin
  Result := FEnv;
end;

procedure TClosure.SetEnv(Value: IEnvironment);
begin
  FEnv := Value;
end;

{ TDefineFunc }

constructor TDefineFunc.Create(const AParamName: string; ABody: IExpression);
begin
  inherited Create;
  FParamName := AParamName;
  FExprs.Add(ABody);
end;

function TDefineFunc.Evaluate(Env: IEnvironment): IExpression;
begin
  Result := TClosure.Create(FExprs[0], Env, FParamName);
end;

function TDefineFunc.GetAsString: string;
begin
  Result := Format('(%s %s %s)', [Self.ClassName, FParamName,
    FExprs[0].AsString]);
end;

{ TCallFunc }

constructor TCallFunc.Create(const AFuncName: string; AParamValue: IExpression);
begin
  inherited Create;
  FFuncName := AFuncName;
  FExprs.Add(AParamValue);
end;

function TCallFunc.Evaluate(Env: IEnvironment): IExpression;
var
  FuncExpr, ParamVal: IExpression;
  Closure: IClosure;
begin
  ParamVal := FExprs[0].Evaluate(Env);
  FuncExpr := Env.GetValue(FFuncName);

  if Supports(FuncExpr, IClosure, Closure) then
    Result := Closure.EvaluateClosure(ParamVal)
  else
    raise EExprException.Create('Invalid expression applied to TCallFunc');
end;

function TCallFunc.GetAsString: string;
begin
  Result := Format('(%s %s)', [FFuncName, FExprs[0].AsString]);
end;

end.
